/*

  logftmp() -- tricks to find out how many anonymous FTP sessoins are going on.

 * Copyright (c) 1988 The Regents of the University of California.
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms are permitted
 * provided that the above copyright notice and this paragraph are
 * duplicated in all such forms and that any documentation,
 * advertising materials, and other materials related to such
 * distribution and use acknowledge that the software was developed
 * by the University of California, Berkeley.  The name of the
 * University may not be used to endorse or promote products derived
 * from this software without specific prior written permission.
 * THIS SOFTWARE IS PROVIDED ``AS IS'' AND WITHOUT ANY EXPRESS OR
 * IMPLIED WARRANTIES, INCLUDING, WITHOUT LIMITATION, THE IMPLIED
 * WARRANTIES OF MERCHANTIBILITY AND FITNESS FOR A PARTICULAR PURPOSE.
 *
 * static char sccsid[] = "@(#)logwtmp.c	5.2 (Berkeley) 9/20/88";
 */

#ifndef lint
static char sccsid[] = "@(#)logftmp.c	5.2 (Berkeley) 9/22/88";
#endif /* not lint */

/* #define USE_FLOCK */

#ifndef  __ARGS
# ifdef __STDC__
#  define __ARGS(x) x
# else
#  define __ARGS(x) ()
# endif
#endif

#include <stdio.h>
#include <unistd.h>
#include <sys/types.h>
#include <fcntl.h>
#include <sys/file.h>
#include <sys/time.h>
#include <sys/stat.h>
#include <syslog.h>
#include "syslog-vals.h"
#include <errno.h>
#include <pwd.h>
#include "custlocality.h"

#ifdef USE_MMAP
# include <sys/mman.h>
#endif

#define	FTMPFILE	"/usr/adm/ftmp"

#define TIMEOUT_SECS	(60*60*24)	/* TWENTY FOUR  hours.. ! */
#define SHORT_TIMEOUT	(60*4)		/* Four minutes */

struct utmp {	/* This is our own utmp... */
	char	ut_name[32];
	char	ut_host[32];
	char	ut_line[12];
	time_t	ut_time;
	unsigned int	locality:8;
	unsigned int	uploader:8;
	unsigned int	sending:1;
	unsigned int	filler1:15;
	time_t  bytecnttime;
	long bytecnt;
	int lastBps;	/* short term average */
	int longBps;	/* long-term average..(whole file) */
	int throttlerate; /* Last value of throttle rate */
	pid_t pid;
	int filler5;
      };

struct summary { /* This is summary data */
	int	filler;
#define FILLER_CODE 0x800004		/* Change when structure, or
					   algorithm is altered... */
	pid_t	server_pid;
	time_t	last_update;		/* Last update of this data */
	int	total_clients;
	long	Bpses[CUST_high+1];	/* Bps     per cust. class */
	long	senders[CUST_high+1];	/* senders per cust. class */
	long	AvgBpses[CUST_high+1];	/* Longer term average */
	long	avgsenders[CUST_high+1];
	int	filler2[86];		/* just in case.. */
};

#ifndef	UTMPDEF_ONLY

static int fd = -1;

#ifdef	USG
#define ftruncate(x,y) {}
#endif

extern int errno;
extern void *malloc __ARGS((unsigned int));
extern time_t time __ARGS((time_t *));
extern void strncpy __ARGS((void *, void *, int));

extern int uploader;	/* ftpd.c main() */

static struct utmp ut;

static int ut_filled = 0;
static struct utmp *uut = 0L;
static int uutsiz = 0;
static time_t next_purgerun = 0;


/* Preparation for MMAP() usage */
static caddr_t memmap = NULL;
static struct summary *summary = NULL;
static struct utmp *umap = NULL;
static int buflen = MAX_FTMP * sizeof(struct utmp) + sizeof(struct summary) + 50;
static int   our_offset = -1;
#ifdef	USE_MMAP
extern	caddr_t mmap( /*caddr_t addr, size_t len, int prot, int flags, int fd, off_t off */);
#endif

static void check_summary_server __ARGS((void));

pid_t mypid;

logftmp(line, name, host, locality, refill, counts)
	char *line, *name, *host;
	int locality, refill;
	int *counts;
{
	int cnt;

	if(fd <0) {
	  fd = open(FTMPFILE, O_RDWR, 0);
	}
	if(fd < 0) return 0;
#ifdef USE_MMAP
	if ( NULL == memmap ) {
	  lseek(fd,buflen,SEEK_SET); /* Write something at the END + 1.. */
	  write(fd,&cnt,1);   /* makes sure there exists space for MAX_FTMP */
	  lseek(fd,0,SEEK_SET);      /* entries and thus avoiding SIGBUS.. */
	  errno = 0;
	  memmap = mmap(/*No fixed place*/NULL, buflen,
			PROT_READ|PROT_WRITE, MAP_SHARED, fd, 0);
	  summary = NULL;
	  if ((long)memmap == -1) {
	    syslog(FTP_LOG_PRIORITY,"logftmp() mmap() failed: %m");
	  } else {
	    /* Get a pointer to the summary structure. */
	    summary = (struct summary*)memmap;
	    /* If UMAP points to something, MMAP() was successfull! */
	    umap    = (struct utmp *)((char *)memmap + sizeof(struct summary));
	  }
	}
#endif
	if( !ut_filled || refill ){
	  ut_filled = 1;
	  mypid = getpid();
	  memset(&ut,0,sizeof(ut));	/* clear it completely at first */
	  ut.pid = mypid;
	  (void)strncpy(ut.ut_line, line, sizeof(ut.ut_line)-1);
	  ut.ut_line[sizeof(ut.ut_line)-1] = 0;
	  (void)strncpy(ut.ut_name, name, sizeof(ut.ut_name)-1);
	  ut.ut_name[sizeof(ut.ut_name)-1] = 0;
	  (void)strncpy(ut.ut_host, host, sizeof(ut.ut_host)-1);	
	  ut.ut_host[sizeof(ut.ut_host)-1] = 0;
	  ut.locality = locality;
	  ut.uploader = uploader;
	}
	ut.sending  = 0;

	(void)time(&ut.ut_time);
	cnt = flock_and_compress(fd,&ut,counts,1/*set our entry*/);
	return cnt;
}

endftmp(line, name, host, locality, counts)
	char *line, *name, *host;
	int locality, *counts;
{
	int cnt;
	extern int uploader;

	if (fd < 0) return 0;
	if( !ut_filled ){
	  ut_filled = 1;
	  mypid = getpid();
	  ut.pid = mypid;
	  (void)strncpy(ut.ut_line, line, sizeof(ut.ut_line));
	  ut.ut_line[sizeof(ut.ut_line)-1] = 0;
	  (void)strncpy(ut.ut_name, name, sizeof(ut.ut_name));
	  ut.ut_name[sizeof(ut.ut_name)-1] = 0;
	  (void)strncpy(ut.ut_host, host, sizeof(ut.ut_host));	
	  ut.ut_host[sizeof(ut.ut_host)-1] = 0;
	  ut.locality = locality;
	  ut.uploader = uploader;
	}
	ut.sending  = 0;
	(void)time(&ut.ut_time);
	cnt = flock_and_compress(fd,&ut,counts,0/*clear our entry*/);
	close(fd);
	fd = -1;

	return cnt;
}

static
#ifdef __GNUC__
inline
#endif
int utmpcmp(up1,up2)
     struct utmp *up1,*up2;
{
	return (up1->pid - up2->pid);
}


int flock_and_compress(fd,ut,counts,flag)
int fd;
struct utmp *ut;
int counts[];
int flag; /* != 0  -> set our entry */
{
#ifdef	USE_FCNTL
	struct flock fl;
#endif
	struct utmp *uup1,*uup2;
	int uutcnt = 0, cnt = 0, rc;
	struct stat buf;
	time_t now;
	int written = 0;
	int i, latest_free, do_lock;
	int seen_self = 0;

	if (counts)
	  for (i=CUST_low; i<=CUST_high; ++i)
	    counts[i] = 0;

	(void)time(&now);

	do_lock = (our_offset == -1) || (now > next_purgerun);
	while(do_lock) {
	  lseek(fd,0,SEEK_SET);
#ifdef	USE_FCNTL
	  fl.l_type   = F_WRLCK;
	  fl.l_whence = SEEK_SET;
	  fl.l_start  = 0;
	  fl.l_len    = 0;
	  if( fcntl(fd,F_SETLK,(void*)&fl) )
#else
#ifdef USE_FLOCK
	  if( flock(fd,LOCK_EX) )
#else
	  if( lockf(fd,F_LOCK,0))
	      /* block until you get it */
#endif
#endif
	    {
	      if(errno==EACCES
#ifdef	EWOULDBLOCK
		 || errno==EWOULDBLOCK
#endif
		 ) {
		microsleep(1000000);
	      } else {
		syslog(LOG_ERR,"locking of ftmp failed, fd=%d, err=%m",fd);
		close(fd);	/* Failure ! */
		fd = -1;
		return 0;
	      }
	    }
	  else break;
	}

#ifdef	USE_MMAP
	uutcnt = MAX_FTMP;
	uut = umap;
	if (umap) {
	  /* Now we have lock, check the existance of summary processor,
	     if it can't be found, or is too old -- over 30 secs since
	     last update, fork() and become one! */
	  check_summary_server();
	} else
#endif
	  {
	    rc = fstat(fd, &buf);
	    if (rc != 0) {
	      syslog(LOG_ERR,"fstat(ftmp_fd,&buf) failed! fd=%d,  err=%m",fd);
	      fd = -1;
	      lseek(fd,0,SEEK_SET);
#ifdef	USE_FCNTL
	      fl.l_type   = F_UNLCK;
	      fl.l_whence = SEEK_SET;
	      fl.l_start  = 0;
	      fl.l_len    = 0;
	      if( fcntl(fd,F_SETLK,(void*)&fl) )
#else
#ifdef USE_FLOCK
	      if (do_lock)
		flock(fd,LOCK_UN);
#else
	      if (do_lock)
		lockf(fd,F_ULOCK,0);
#endif
#endif
	      return 0;
	    }
	    uutcnt = buf.st_size / sizeof(struct utmp);
	    if (!uut) {
	      uut = (struct utmp *)malloc(buf.st_size+1+sizeof(struct utmp));
	      uutsiz = buf.st_size+1+sizeof(struct utmp);
	    }
	    if (uutsiz < buf.st_size) {
	      uut = (struct utmp *)realloc(uut,buf.st_size+sizeof(struct utmp));
	      uutsiz = buf.st_size+sizeof(struct utmp);
	    }
	    if (buf.st_size > 0)
	      read(fd, (void*)uut, buf.st_size);
	  }
	/* process it! */
	uup2 = uut;
	cnt = 0;
	latest_free = -1;
	--uup2; /* one before.. */
	for(i=0; i<uutcnt; ++i) {
	  ++uup2;
	  ++cnt;
	  if (utmpcmp(uup2,ut)==0) {
	    /* Thats ours.. */
	    our_offset = i * sizeof(struct utmp);
	    seen_self = 1;
	    continue;
	  }
	  if (uup2->pid == 0) {
	    /* Empty slot! */
	    --cnt;
	    if (latest_free < 0)
	      latest_free = i;
	    continue;
	  }
	  if (!uup2->uploader && counts)
	    counts[uup2->locality] += 1;
	}
	if (!seen_self)
	  our_offset = -1;
	if (our_offset == -1 /*&& do_lock*/) {
	  /* Store our own entry into a free slot */
	  /* as long as there is known free slot.. */
	  if (latest_free >= 0)
	    our_offset = latest_free * sizeof(struct utmp);
	  else
	    our_offset = -1;
	}
	if (flag) { /* set our entry */
	  if (our_offset >= 0)
	    memcpy((char*)uut+our_offset,ut,sizeof(struct utmp));
	} else { /* clear our entry */
	  if (our_offset >= 0) {
	    uup2=(struct utmp *)((char*)uut+our_offset);
	    if (uup2->pid == 0)
	      syslog(FTP_LOG_PRIORITY,
		     "purging own entry: '%s', '%s', '%s', %d",
		     uup2->ut_name,uup2->ut_host,uup2->ut_line,uup2->pid);
	    memset((char*)uut+our_offset,0,sizeof(struct utmp));
	    our_offset = -1;
	  }
	}
	/* if (our_offset < 0)
	   syslog(FTP_LOG_PRIORITY,"our_offset < 0!"); */
	if (!seen_self && !do_lock)
	  syslog(FTP_LOG_PRIORITY,"missing permit to %s own entry!  latest_free=%d",
		 flag?"set":"clear", latest_free);
	written = 1;
	
	/* output it */

	if ( !umap ) {		/* No MMAP() buffer	*/
	  lseek(fd, 0, SEEK_SET);	/* Can't fail ??	*/
	  ftruncate(fd, buflen+1);
	  if (uutcnt > 0)
	    write(fd, (void*)uut, uutcnt*sizeof(struct utmp));
	}

	/*free(uut);*/

	if (now > next_purgerun)
	  next_purgerun = now+4*60; /* Four minutes apart.. */


	lseek(fd,0,SEEK_SET);
#ifdef	USE_FCNTL
	fl.l_type   = F_UNLCK;
	fl.l_whence = SEEK_SET;
	fl.l_start  = 0;
	fl.l_len    = 0;
	if( fcntl(fd,F_SETLK,(void*)&fl) )
#else
#ifdef USE_FLOCK
	if (do_lock)
	  flock(fd,LOCK_UN);
#else
	if (do_lock)
	  lockf(fd,F_ULOCK,0);
#endif
#endif
	return cnt;
}


int getsetftmpBps(sending,mybytecnt,mybytecnttime,longBps,myBps,throttlerate,Bpsmyclass,myclasssenders)
  int sending, mybytecnt, longBps, myBps, throttlerate, *Bpsmyclass,*myclasssenders;
  time_t mybytecnttime;
{
	struct utmp *uup2, *uup, ut2;
	int uutcnt = 0;
	struct stat buf;
	int i;
	int rc;
	time_t now;


	if (our_offset < 0) return 0; /* No place is known :-( */

	(void)time(&now);
	/* our_offset is known with & without mmap() */
#ifdef	USE_MMAP
	uutcnt = MAX_FTMP;
	uut = umap;
	uup = (struct utmp *)((long)uut + our_offset);
	if ( !umap )
#endif
	  {
	    lseek(fd,0,SEEK_SET);
	    while(1) {		/* block until you get it */
#ifdef	USE_FCNTL
	      struct flock fl;
	      fl.l_type   = F_WRLCK;
	      fl.l_whence = SEEK_SET;
	      fl.l_start  = 0;
	      fl.l_len    = 0;
	      if( fcntl(fd,F_SETLK,(void*)&fl) )
#else
#ifdef USE_FLOCK
	      if( flock(fd,LOCK_EX) )
#else
		if( lockf(fd,F_LOCK,0))
#endif
#endif
		  {
		    if(errno==EACCES
#ifdef	EWOULDBLOCK
		       || errno==EWOULDBLOCK
#endif
		       ) {
		      microsleep(1000000);
		    } else {
		      syslog(LOG_ERR,"locking of ftmp failed, fd=%d, err=%m",fd);
		      close(fd); /* Failure ! */
		      fd = -1;
		      return 0;
		    }
		  }
		else break;
	    }

	    rc = fstat(fd, &buf); /* Can't fail ??		*/
	    if (rc != 0) {
	      syslog(LOG_ERR,"fstat(ftmp_fd,&buf) failed! fd=%d,  err=%m",fd);
	      fd = -1;
	      return 0;
	    }
	    uutcnt = buf.st_size / sizeof(struct utmp);

	    if (!uut) {
	      uut = (struct utmp *)malloc(buf.st_size+1);
	      uutsiz = buf.st_size+1;
	    }
	    if (uutsiz < buf.st_size) {
	      uut = (struct utmp *)realloc(uut,buf.st_size);
	      uutsiz = buf.st_size;
	    }
	    uup = (struct utmp *)((long)uut + our_offset);
	    if (buf.st_size > 0)
	      read(fd, (void*)uut, buf.st_size);
	  }

	/* process it! */

	if (Bpsmyclass)
	  *Bpsmyclass = 0;
	if (myclasssenders)
	  *myclasssenders = 0;
	uup2 = uut;
	rc = -1;
	--uup2;
	/* If pointer to OUR dataslot exists, update our data! */
	if (uup) {
	  /* Thats ours.. */
	  uup->bytecnt = mybytecnt;
	  uup->bytecnttime = mybytecnttime;
	  uup->lastBps = myBps;
	  uup->sending = sending;
	  uup->longBps = longBps;
	  uup->throttlerate = throttlerate;
	  rc = 0;
	}

	if (!Bpsmyclass || !myclasssenders) {
	  /* If no pointers to return values, just update
	     data in our slot; -- did it above -- if such
	     slot (uup->) exists. */
	} else {
	  /* Right; return-value pointers exist,
	     now if there exists summary structure.. */
	  if (summary && uup) {
	    /* All quick-access pointers exist */
	    if (Bpsmyclass)
	      *Bpsmyclass = summary->Bpses[uup->locality];
	    if (myclasssenders)
	      *myclasssenders = summary->senders[uup->locality];
	  } else {
	    /* No, no summary, just scan thru like before.. */
	    for(i=0; i<uutcnt; ++i) {
	      ++uup2;
	      if (uup2 == uup) {
		/* Thats ours.. */
		uup2->bytecnt = mybytecnt;
		uup2->bytecnttime = mybytecnttime;
		uup2->lastBps = myBps;
		uup2->sending = sending;
		uup2->longBps = longBps;
		uup2->throttlerate = throttlerate;
		rc = 0;
	      }
	      if ((ut.locality == uup2->locality) &&
		  uup2->sending &&
		  ((uup2->bytecnttime + SHORT_TIMEOUT) > now)) {
		/* Same locality, and not ages old.. */
		if (Bpsmyclass)
		  *Bpsmyclass += uup2->lastBps;
		if (myclasssenders)
		  *myclasssenders += 1;
	      }
	    }
	  }
	}
	/* output it */
	if (0 == rc && our_offset >= 0 && !umap) {
	  lseek(fd, our_offset, SEEK_SET);
	  write(fd, (void*)uup, buf.st_size);
	}
#if 0 /* No more  needed... (??) */
	else {   /* Our entry has been deleted ?  Well, put it there again.*/
	  memcpy(&ut2,&ut,sizeof(ut2));
	  ut2.bytecnt = mybytecnt;
	  ut2.bytecnttime = mybytecnttime;
	  ut2.lastBps = myBps;
	  ut2.sending = sending;
	  ut2.longBps = longBps;
	  ut2.throttlerate = throttlerate;
	  lseek(fd, buf.st_size, SEEK_SET);
	  write(fd, &ut2, sizeof(struct utmp));
	  *myclasssenders += 1;
	}
#endif
	/*free(uut);*/
	if ( !umap ) {
#ifdef	USE_FCNTL
	  struct flock fl;
	  fl.l_type   = F_UNLCK;
	  fl.l_whence = SEEK_SET;
	  fl.l_start  = 0;
	  fl.l_len    = 0;
	  fcntl(fd,F_SETLK,(void*)&fl);
#else
#ifdef USE_FLOCK
	  lseek(fd,0,SEEK_SET);
	  flock(fd,LOCK_UN);
#else
	  lseek(fd,0,SEEK_SET);
	  lockf(fd,F_ULOCK,0);
#endif
#endif
	}
	return 0;
}



int ftmptrunc()
{
	extern struct passwd *pw;	

	if (pw == NULL || pw->pw_uid != 0)
	  return -1;
	/* If not logged in, or not root-capable, don't do it.. */

	lseek(fd,0,SEEK_SET);

	while(1) {	    /* block until you get it */
#ifdef	USE_FCNTL
	  struct flock fl;
	  fl.l_type   = F_WRLCK;
	  fl.l_whence = SEEK_SET;
	  fl.l_start  = 0;
	  fl.l_len    = 0;
	  if (fcntl(fd,F_SETLK,(void*)&fl) )
#else
#ifdef USE_FLOCK
	  if( flock(fd,LOCK_EX) )
#else
	  if( lockf(fd,F_LOCK,0))
#endif
#endif
	    {
	      if(errno==EACCES
#ifdef	EWOULDBLOCK
		 || errno==EWOULDBLOCK
#endif
		 ) {
		microsleep(1000000);
	      } else {
		close(fd);	/* Failure ! */
		fd = -1;
		return 0;
	      }
	    }
	  else break;
	}

	if ( umap ) {
	  memset((void*)memmap, 0, buflen);
	} else {
	  lseek(fd, 0, SEEK_SET);
	  ftruncate(fd,0);
	}

	/*fsync(fd);*/
#ifdef	USE_FCNTL
	{
	  struct flock fl;
	  fl.l_type   = F_UNLCK;
	  fl.l_whence = SEEK_SET;
	  fl.l_start  = 0;
	  fl.l_len    = 0;
	  fcntl(fd,F_SETLK,(void*)&fl);
	}
#else
	lseek(fd,0,SEEK_SET);
#ifdef USE_FLOCK
	flock(fd,LOCK_UN);
#else
	lockf(fd,F_ULOCK,0);
#endif
#endif
	return 0;
}


/* ================================================================ */

static void
summary_server()
{
	/* Summary server:
	   This runs a loop every 5-10 seconds (or so), and calculates what
	   the Bpses and senders is to be for every user class.
	   */
	struct utmp *utp;
	int i, cnt;
	struct summary sum;
	int once = 1;
	time_t now;

	memset(&sum,0,sizeof(sum));

	sum.filler = FILLER_CODE; /* Change this when contents change.. */
	sum.server_pid = getpid();
	for (;;) {
	  if (sum.server_pid != getpid()) {
	    exit(1); /* Ok, kill myself if no data can be entered.. */
	  }
	  time(&now);
	  for (i=CUST_low;i<=CUST_high;++i) {
	    sum.Bpses[i] = 0;
	    sum.senders[i] = 0;
	  }
	  utp = uut;
	  --utp;
	  cnt = 0;
	  for (i=0; i<MAX_FTMP; ++i) {
	    ++utp;
	    ++cnt;
	    if (utp->pid == 0) {
	      /* Empty slot! */
	      --cnt;
	      continue;
	    }
	    if (now > next_purgerun &&
		kill(utp->pid,0) != 0) {
	      syslog(FTP_LOG_PRIORITY,
		     "summary_server(): purging nonexistent: '%s', '%s', '%s', %d",
		     utp->ut_name,utp->ut_host,utp->ut_line,utp->pid);
	      memset(utp, 0, sizeof(struct utmp));
	      --cnt;
	      continue;
	    }
	    if ((now > next_purgerun) &&
		((now > (utp->ut_time + TIMEOUT_SECS) &&
		  utp->bytecnttime != 0 &&
		  now > (utp->bytecnttime + SHORT_TIMEOUT*2)) ||
		 (utp->bytecnttime == 0 &&
		  now > (utp->ut_time + TIMEOUT_SECS)))) {
	      /* This bastard exists but has not had any xfer activity for a
		 while -- kill(pid,0) returned ok */
	      kill(utp->pid,9);
	      syslog(FTP_LOG_PRIORITY,
		     "summary_server() killed frozen FTP server process: '%s', '%s', '%s', Age: %ds, %d",
		     utp->ut_name,utp->ut_host,utp->ut_line,
		     utp->ut_time - now, utp->pid);
	      /* Do NOT zap him yet! 
		 memset(utp, 0, sizeof(struct utmp)); */
	      --cnt;		/* That was TIMEOUT_SECS old, updated at every
				   retrieve, and during data transport */
	      continue;		/* Outdated!    */
	    }
	    if (utp->sending && (utp->bytecnttime + SHORT_TIMEOUT) > now) {
	      sum.senders[utp->locality] += 1;
	      sum.Bpses  [utp->locality] += utp->lastBps;
	    }
	  }
	  sum.total_clients = cnt;
	  time(&sum.last_update);

	  for (i=CUST_low;i<=CUST_high;++i) {
	    sum.AvgBpses[i] = (95*sum.AvgBpses[i])/100+5*sum.Bpses[i];
	    sum.avgsenders[i] = (95*sum.avgsenders[i])/100+5*sum.senders[i];
	  }

	  memcpy(summary,&sum,sizeof(sum));

	  if (once)
	    syslog(FTP_LOG_PRIORITY,"summary_server() started, pid=%d",
		   getpid());
	  once = 0;
	  if (now > next_purgerun)
	    next_purgerun = now+3*60; /* Three minutes apart.. */
	  microsleep(1000000); /* Once in a second.. */
	} /* loop forever */
}


static void
check_summary_server()
{
	int rc;
	int pid;
	time_t now;

	time(&now);

	if (!summary) /* No pointer there, propably not in MMAP mode.. */
	  return;

	rc = kill(summary->server_pid,0);
	if (summary->server_pid < 100 || rc != 0 ||
	    now > (summary->last_update+30)) {
	  /* No summary server ?  Become one! */
	  if (summary->server_pid >= 100 && rc == 0 &&
	      now > (summary->last_update+30)) {
	    /* There is a process, it exists, it has been far too lazy!
	       KILL IT        */
	    kill(summary->server_pid, 9);
	    syslog(FTP_LOG_PRIORITY,
		   "check_summary_server() killed frozen server process: pid %d, Age %ds",
		   summary->server_pid,now - summary->last_update);
	    summary->server_pid = 0; /* No process anymore.. */
	    summary->last_update = 0;
	  }
	  pid = fork();
	  if (pid == 0) { /* Child! */
#ifdef	SETPROCTITLE
	    setproctitle("FTPD: summary_server()");
#endif
	    /* Close everything */
	    for (rc=getdtablesize()-1; rc >= 0; --rc)
	      close(rc);
	    summary_server();
	  }
	  microsleep(2000000 /* 2 seconds */);
	}
}

#endif /* UTMPDEF_ONLY */
